@@ -1,5 +1,7 @@
 package RFID::Matrics::Reader;
-our $VERSION = '0.001';
+our $VERSION = '0.002';
+@ISA = qw(RFID::Reader Exporter);
+use strict;
 
 # Written by Scott Gifford <gifford@umich.edu>
 # Copyright (C) 2004 The Regents of the University of Michigan.
@@ -14,22 +16,28 @@ RFID::Matrics::Reader - Abstract base class for a Matrics RFID reader
 
 This abstract base class provides most of the methods required for
 interfacing Perl with a Matrics RFID reader.  To actually create an
-object, use L<RFID::Matrics::Reader::Serial> or
-L<RFID::Matrics::Reader::TCP>.
+object, use
+L<RFID::Matrics::Reader::Serial|RFID::Matrics::Reader::Serial> or
+L<RFID::Matrics::Reader::TCP|RFID::Matrics::Reader::TCP>.  It is based
+on L<RFID::Reader|RFID::Reader>.
 
     use RFID::Matrics::Reader::Serial;
     my $reader = 
-      RFID::Matrics::Reader::Serial->new(comport => $com,
-				         node => 4,
-					 antenna => MATRICS_ANT_1);
-    $reader->changeparam(antenna => MATRICS_ANT_1,
-			 environment => 4,
-			 power_level => 0xff,
-			 combine_antenna_bits => 0);
-    my $rff = $reader->readfullfield_unique(antenna => MATRICS_ANT_1);
-    foreach my $tag (@{$pp->{utags}})
+      RFID::Matrics::Reader::Serial->new(Port => $com_port_object,
+				         Node => 4,
+					 Antenna => 1)
+        or die "Couldn't create reader object\n";
+
+    my @err = $reader->set(PowerLevel => 0xff,
+		           Environment => 4) == 0
+        or die "Couldn't set params: @err\n";
+
+    my @tags = $reader->readtags;
+    foreach my $tag (@tags)
     {
-	print "I see tag $tag->{id}\n";
+	my $tag_info = $tag->get('Antenna','ID','Type');
+	print "I see tag $tag_info{Type}.$tag_info{ID} ".
+              "at antenna $tag_info{Antenna}.\n";
     }
 
 =head1 DESCRIPTION
@@ -41,38 +49,99 @@ Specification>, using version 2.8 from October 19th 2003.  It was
 tested with an RDR-001 model reader.
 
 To actually create a reader object, use
-L<RFID::Matrics::Reader::Serial> or L<RFID::Matrics::Reader::TCP>.
-Those classes inherit from this one.
+L<RFID::Matrics::Reader::Serial|RFID::Matrics::Reader::Serial> or
+L<RFID::Matrics::Reader::TCP|RFID::Matrics::Reader::TCP>.  Those
+classes inherit from this one.
+
+This class inherits some methods and settings from
+L<RFID::Reader|RFID::Reader>.
 
 =cut
 
+use RFID::Reader qw(hexdump);
 use Exporter;
 use Carp qw(cluck croak carp);
 use vars qw(@ISA @EXPORT_OK %EXPORT_TAGS);
-@ISA = qw(Exporter);
+# You shouldn't use the antenna constants anymore, just use numbers 1-4.
 @EXPORT_OK = qw(MATRICS_ANT_1 MATRICS_ANT_2 MATRICS_ANT_3 MATRICS_ANT_4 hexdump);
 %EXPORT_TAGS = (ant => [ qw(MATRICS_ANT_1 MATRICS_ANT_2 MATRICS_ANT_3 MATRICS_ANT_4)]);
 
 use RFID::Matrics::CRC qw(crc);
-use RFID::Matrics::Tag qw(:tagtypes tagcmp);
-
-=head2 CONSTANTS
-
-=over 4
-
-=item MATRICS_ANT_I<n>
-
-Specific antennas on the reader.  The constants I<MATRICS_ANT_1>,
-I<MATRICS_ANT_2>, I<MATRICS_ANT_3>, I<MATRICS_ANT_4> are provided.
-
-=cut
+use RFID::EPC::Tag;
+use RFID::Matrics::Tag;
 
 use constant MATRICS_ANT_1 => 0xa0;
 use constant MATRICS_ANT_2 => 0xb0;
 use constant MATRICS_ANT_3 => 0xc0;
 use constant MATRICS_ANT_4 => 0xd0;
 
-=back
+our %antname = (
+		&MATRICS_ANT_1 => 1,
+		&MATRICS_ANT_2 => 2,
+		&MATRICS_ANT_3 => 3,
+		&MATRICS_ANT_4 => 4,
+		);
+
+our %paramblock_setting = (power_level => 1,
+                           environment => 1,
+                          );
+our %paramblock_rename = (PowerLevel => 'power_level',
+                          Environment => 'environment',
+                          Mask => 'Mask',
+                         );
+our %readerstatus_rename = (ReaderVersion => 'version',
+                            ReaderSerialNum => 'serialnum',
+                           );
+
+our %antid = (1 => MATRICS_ANT_1,
+              2 => MATRICS_ANT_2,
+              3 => MATRICS_ANT_3,
+              4 => MATRICS_ANT_4);
+
+# Initializer used by derived objects
+sub _init
+{
+    my $self = shift;
+    my(%p)=@_;
+
+    $self->{default_node} = $p{Node}||$p{node};
+    $self->{timeout} = $p{timeout}||$p{Timeout}
+        unless ($self->{timeout});
+
+    # This can't be set again, so let's delete it.
+    delete $p{node};
+    delete $p{Node};
+
+    my @errs = $self->set(%p);
+    if (grep { $_ !~ /Unknown setting/ } @errs)
+    {
+	die "Error setting properties: @errs\n";
+    }
+
+    $self->{default_antenna} = $p{Antenna}||$p{antenna};
+    if (!defined($self->{default_antenna}))
+    {
+	 if (defined($self->{_antenna_sequence}))
+	 {
+	     $self->{default_antenna} = $self->{_antenna_sequence}[0];
+	 }
+	 else
+	 {
+	     $self->{default_antenna} = 1;
+	 }
+    }
+    $self->{_antenna_sequence} = [$self->{default_antenna}]
+	if (!$self->{_antenna_sequence});
+
+
+    $self->stop_constant_read(node => $self->{default_node})
+	if ($self->{default_node} and (!$p{noinit}));
+
+    $self->SUPER::_init(%p);
+}
+
+
+
 
 =head2 METHODS
 
@@ -93,21 +162,6 @@ our %_errmsgs = (
 # Prototype
 sub uniq(&@);
 
-# Initializer used by derived objects
-sub _init
-{
-    my $self = shift;
-    my(%p)=@_;
-
-    $self->{default_node} = $p{node};
-    $self->{default_antenna} = $p{antenna};
-    $self->{debug} = $p{debug};
-    $self->{timeout} = $p{timeout}||$p{Timeout};
-
-    $self->stop_constant_read(node => $self->{default_node})
-	if ($self->{default_node} and (!$p{noinit}));
-}
-
 sub _setdefaults
 {
     my $self = shift;
@@ -116,6 +170,8 @@ sub _setdefaults
       unless defined($p->{node});
     $p->{antenna} = $self->{default_antenna}
       unless defined($p->{antenna});
+    $p->{antenna} = $antid{$p->{antenna}}
+      if (defined($p->{antenna}) and $p->{antenna} < 10);
     $p;
 }
 
@@ -124,7 +180,7 @@ sub _makepacket
     my $self = shift;
     my(%p)=@_;
     $self->_setdefaults(\%p);
-
+    
     $p{data}||="";
     my $packet = pack("CCCa*",$p{node},length($p{data})+5,$p{cmd},$p{data});
     return pack("Ca*v",1,$packet,crc($packet));
@@ -158,19 +214,19 @@ sub _parsepacket
 sub _getpacket
 {
     my($self)=@_;
-    $data = $self->_readbytes(3)
-	or die "Couldn't read COM port: $!\n";
+    my $data = $self->_readbytes(3)
+	or die "Couldn't read data: $!\n";
     length($data) == 3
-	or die "COM port short read!\n";
+	or die "Data short read!\n";
     my($sof,$addr,$len)=unpack("CCC",$data);
     
     my $moredata = $self->_readbytes($len-2)
-	or die "Couldn't read COM port: $!\n";
+	or die "Couldn't read data: $!\n";
     length($moredata) == ($len-2)
-	or die "COM port short read!\n";
+	or die "Data short read!\n";
     
-    warn time()," RECV: ",hexdump($data.$moredata),"\n"
-	if ($self->{debug});
+    $self->debug(" RECV: ",hexdump($data.$moredata),"\n")
+	if ($self->{_debug});
     return $data.$moredata;
 }
 
@@ -179,8 +235,8 @@ sub _sendpacket
     my $self = shift;
     my($data)=@_;
 
-    warn time()," SEND: ",hexdump($data),"\n"
-	if ($self->{debug});
+    $self->debug(" SEND: ",hexdump($data),"\n")
+	if ($self->{_debug});
     $self->_writebytes($data)
 	or die "Couldn't write to COM port: $^E";
 }
@@ -208,54 +264,115 @@ sub _do_something
     return wantarray?@ret:$ret[0];
 }
 
+=head3 get
 
-=head3 getparamblock
-
-Gets the configuration for a particular antenna on a particular
-reader.  Takes a hash as an argument, which can have the following
-parameters:
-
-=over 4
-
-=item node
-
-Node to be queried.
-
-=item antenna
-
-Antenna whose configuration you want
-
-=back
-
-Returns a hash reference with the following attributes:
-
-=over 4
-
-=item power_level
-
-The amount of power this antenna should use when doing a read, between
-0 and 255.  255 is full-power; the scale of this setting is
-logarithmic, so 0xC0 is about 50% power, and 0x80 is about 25% power.
-
-=item environment
-
-How long the reader should try to read tags during a C<readfullfield>
-command, between 0 and 4.  0 will read for a very short time, and is
-appropriate for environments where tags come and go very quickly, and
-it's OK if you miss a tag somtimes.  4 will read for longer, and is
-appropriate where tags stay relatively static and you want the reader
-to try its best to find all of them.
-
-=back
-
-The parameters for combined antennas and filters are not yet fully
-implemented.
+Get various properties of the reader or the internal state of the
+object.  The syntax is described in L<the RFID::Reader get
+method|RFID::Reader/get> documentation.  See L<Matrics Properties|/Properties> 
+and L<Generic Properties|RFID::Reader/Properties> for the properties
+that can be retreived.
 
 =cut
 
+sub get
+{
+    my $self = shift;
+    my %get;
+    my %ret;
+    my %paramblocks;
+    my $readerstatus;
+
+    foreach my $g (@_)
+    {
+	if ($paramblock_rename{$g} or ($g =~ /^(\w+)_Antenna(\d+)$/ and $paramblock_rename{$1}))
+	{
+	    my($ant,$asv);
+	    if ($paramblock_rename{$g})
+	    {
+		$ant = $self->{default_antenna};
+		$asv = $paramblock_rename{$g};
+	    }
+	    else
+	    {
+		$ant = $2;
+		$asv = $paramblock_rename{$1};
+	    }
+	    if (!$paramblocks{$ant})
+	    {
+		$paramblocks{$ant} = $self->getparamblock(antenna => $antid{$ant})
+		    or die "Couldn't getparamblock for antenna $ant!";
+	    }
+	    if ($g eq 'Mask')
+	    {
+		my $masktype = $paramblocks{$ant}{filter_type};
+		if ($masktype == 0)
+		{
+		    $ret{$g}='';
+		}
+		else
+		{
+		    my $mask = bin2hex_big_endian($paramblocks{$ant}{filter_bits});
+		    $mask =~ s/\s//g;
+		    if ($masktype == 1)
+		    {
+			$ret{$g} = substr($mask,0,2)."/8";
+		    }
+		    elsif ($masktype == 2)
+		    {
+			$ret{$g} = substr($mask,0,10)."/40";
+		    }
+		    elsif ($masktype == 3)
+		    {
+			$ret{g} = substr($mask,0,16)."/64";
+		    }
+		    else
+		    {
+			# Error
+			$ret{g} = undef;
+		    }
+		}
+	    }
+	    else
+	    {
+		$ret{$g} = $paramblocks{$ant}{$asv};
+	    }
+	}
+	elsif ($g eq 'AntennaSequence')
+	{
+	    $ret{$g} = [$self->{_antenna_sequence}];
+	}
+	elsif ($readerstatus_rename{$g})
+	{
+	    if (!$readerstatus)
+	    {
+		my %p;
+		$self->_setdefaults(\%p);
+		$readerstatus = $self->getreaderstatus(%p)
+		    or die "Couldn't get reader status: $self->{error}";
+	    }
+	    $ret{$g}=$readerstatus->{$readerstatus_rename{$g}};
+	}
+	else
+	{
+	    croak "Unknown setting '$g'";
+	}
+    }
+    if (wantarray)
+    {
+	return %ret;
+    }
+    else
+    {
+	# Return last value
+	return $ret{$_[$#_]};
+    }
+
+}
+
 sub getparamblock
 {
     my $self = shift;
+    my(%p)=@_;
     $self->_setdefaults(\%p);
     $self->_do_something(\&_cmd_getparamblock,\&_resp_getparamblock,@_);
 }
@@ -277,7 +394,7 @@ sub _resp_getparamblock
     my $self = shift;
     my $pp = $self->_parsepacket(@_)
 	or return undef;
-    (@$pp{qw(power_level environment combine_antenna_bits protocol_speed filter_len tagtype reserved_bits filter_bits reserved_bits)}) =
+    (@$pp{qw(power_level environment combine_antenna_bits protocol_speed filter_type tagtype reserved_bits filter_bits reserved_bits)}) =
 	unpack("CCa1CCCa2a8a*",$pp->{data});
     $pp->{combine_antenna}=[];
     my $ca = ord $pp->{combine_antenna_bits};
@@ -286,22 +403,154 @@ sub _resp_getparamblock
 	my @antarr = (MATRICS_ANT_1, MATRICS_ANT_2, MATRICS_ANT_3, MATRICS_ANT_4);
 	if ($ca & (1 << $i))
 	{
-	    push(@{$pp{combine_antenna}},$antarr[$i]);
+	    push(@{$pp->{combine_antenna}},$antarr[$i]);
 	}
     }
     $pp;
 }
 
-=head3 setparamblock
 
-Sets parameters for a specific reader and antenna.  It can set all of
-the parameters which can be read by I<getparamblock>.
+=head3 set
+
+Set one or more properties associated with a reader.  Depending on
+implementation, this may send one or more commands to the reader, set
+an internal flag, or take some other action.  
+
+This method takes a hash with the properties to be set as keys, and
+their new values as values.  It returns a list of errors that occured;
+if no errors occured, it will return an empty list. In a scalar
+context, that evaluates to the number of errors that occured, so you
+can test for errors like this:
+
+    my @errs = $reader->set(SomeVariable => "New Value") == 0
+      or die "Couldn't set SomeVariable: @errs";
+
+See L<Properties|/Properties> for the properties that can be set.
 
 =cut
 
+sub set
+{
+    my $self = shift;
+    my(%p)=@_;
+    my %antset;
+    my %ant_specific;
+    my %unknown;
+    my @errs;
+
+    # First pass through settings.
+    # Settings that can be grouped into a single command are collected,
+    # and set efficiently in a second pass.
+    while(my($k,$v)=each(%p))
+    {
+	if ($paramblock_rename{$k} or ($k =~ /^(\w+)_Antenna(\d+)$/ and $paramblock_rename{$1}))
+	{
+	    my($ant,$param);
+	    if ($paramblock_rename{$k})
+	    {
+		($ant,$param)=('ALL',$paramblock_rename{$k});
+	    }
+	    else
+	    {
+		($ant,$param)=($2,$paramblock_rename{$1});
+	    }
+	    if ($param eq 'Mask')
+	    {
+		if ($v eq '')
+		{
+		    $antset{$ant}{filter_type}=0;
+		    $antset{$ant}{filter_bits}="\0"x8;
+		}
+		else
+		{
+		    my($mask,$len,$start)=split(/\//,$v);
+		    if ($start)
+		    {
+			push(@errs,"Matrics reader doesn't support mask start bit");
+			next;
+		    }
+		    if (!defined($len))
+		    {
+			$len = length($mask) * 4;
+		    }
+		    $antset{$ant}{filter_bits} = hex2bin_big_endian($mask);
+		    if ($len == 0)
+		    {
+			$antset{$ant}{filter_type} = 0;
+		    }
+		    elsif ($len == 8)
+		    {
+			$antset{$ant}{filter_type} = 1;
+		    }
+		    elsif ($len == 40)
+		    {
+			$antset{$ant}{filter_type} = 2;
+		    }
+		    elsif ($len == 64)
+		    {
+			$antset{$ant}{filter_type} = 3;
+		    }
+		    else
+		    {
+			push(@errs,"Matrics reader only supports mask len of 0, 8, 40, and 64.");
+			next;
+		    }
+		    $antset{$ant}{filter_bits} = substr($antset{$ant}{filter_bits},0,$len/8) . "\0" x (8-($len/8));
+		}
+	    }
+	    $antset{$ant}{$param} = $v;
+	}
+	elsif ($k eq 'AntennaSequence')
+	{
+	    $self->{_antenna_sequence} = $v;
+	}
+	elsif ($k eq 'node')
+	{
+	    push(@errs, "Can't set node (yet)!");
+	    next;
+	}
+	elsif ($k eq 'baudrate')
+	{
+	    push(@errs,"Can't set baudrate (yet)!");
+	    next;
+	}
+	else
+	{
+	    $unknown{$k}=$v;
+	}
+    }
+    if (keys %unknown)
+    {
+	push(@errs,$self->SUPER::set(%unknown));
+    }
+
+    # Copy options for all antennas into each individual antenna
+    while (my($k,$v)=each(%{$antset{ALL}}))
+    {
+	foreach my $a (@{$self->{_antenna_sequence}})
+	{
+	    $antset{$a}{$k} = $v;
+	}
+    }
+
+    foreach my $ant (1..4)
+    {
+	if ($antset{$ant})
+	{
+	    $antset{$ant}{antenna} = $antid{$ant};
+	    $self->_setdefaults($antset{$ant});
+	    $self->changeparamblock(%{$antset{$ant}})
+		or push(@errs, "Error changing params: $self->{error}");
+	}
+    }
+    @errs;
+}
+
+
 sub setparamblock
 {
     my $self = shift;
+    my(%p)=@_;
     $self->_setdefaults(\%p);
 
     $self->_do_something(\&_cmd_setparamblock,\&_resp_setparamblock,@_);
@@ -316,11 +565,11 @@ sub _cmd_setparamblock
 		       cmd => 0x23,
 		       data => pack("CCCCCCCCCCx2a8x16",
 				    $self->_make_confwhich_ant(\%p), # Returns 4 bytes
-				    $p{power_level}||0xff,
-				    $p{environment}||0x00,
+				    defined($p{power_level})?$p{power_level}:0xff,
+				    defined($p{environment})?$p{environment}:0x00,
 				    $self->_make_combine_antenna_bits(\%p),
 				    $p{protocol_speed}||0,
-				    $p{filter_len}||0,
+				    $p{filter_type}||0,
 				    $p{tagtype}||0,
 				    $p{filter_bits}||"\0"x8,
 				    )
@@ -334,17 +583,6 @@ sub _resp_setparamblock
 	or return undef;
 }
 
-=head3 changeparamblock
-
-Reads the current parameters from the reader, changes any which are
-specified in the arguments, then writes them back out.  This is
-impelemented in terms of I<getparamblock> and I<setparamblock>, and
-recognizes the same parameters as they do.  Returns a hash reference
-with the following attributes:
-
-
-=cut
-
 sub changeparamblock
 {
     my $self = shift;
@@ -362,29 +600,32 @@ sub changeparamblock
     return $self->setparamblock(%$curparam, @_);
 }
 
-=head3 readfullfield
+=head3 readtags
 
-Read all tags in the field of the specified antenna.  Takes parameters
-I<node> and I<antenna> to specify what field should be read, and
-returns a reference to a hash containing the following attributes:
-
-=over 4
-
-=item numtags
-
-The total number of tags read.  Note that this number can contain
-duplicates; use I<readfullfield_unique> to filter out duplicates.
-
-=item tags
-
-An array reference containing zero or more
-L<RFID::Matrics::Tag|RFID::Matrics::Tag> objects representing the tags
-that were read.
-
-=back
+Read all of the tags in the reader's field, honoring any settings
+affecting the reading and filtering of tags.  This returns a (possibly
+empty) list of tags, which will be of type
+L<RFID::EPC::Tag|RFID::EPC::Tag> or
+L<RFID::Tag::Matrics|RFID::Matrics::Tag>.  See L<the RFID::Reader
+readtags method documentation|RFID::Reader/readtags> for more
+information.
 
 =cut
 
+sub readtags
+{
+    my $self = shift;
+    my @tags;
+    
+    foreach my $ant (@{$self->{_antenna_sequence}})
+    {
+	my $r = $self->readfullfield(antenna => $ant);
+	push(@tags,@{$r->{tags}})
+	    if ($r->{tags});
+    }
+    return $self->filter_tags(@tags);
+}
+
 sub readfullfield
 {
     my $self = shift;
@@ -394,6 +635,7 @@ sub readfullfield
     foreach my $r (@resp)
     {
 	$ret->{numtags} += $r->{numtags};
+	push(@{$ret->{tags}}, @{$r->{tags}});
     }
     $ret;
 }
@@ -417,77 +659,22 @@ sub _resp_readfullfield
 	or return undef;
     my $dc = $pp->{data};
     (@$pp{qw(antenna numtags)}) = unpack("CC",substr($dc,0,2,""));
-    $pp->{tags} = [$self->_parsetags($pp->{numtags},$dc)];
+    $pp->{tags} = [$self->_parsetags($pp->{numtags},$dc,Antenna => $antname{$pp->{antenna}})];
     $pp;
 }
 
-
-=head3 readfullfield_unique
-
-Performs a I<readfullfield>, then adds the following attributes to the
-hash reference:
-
-=over 4
-
-=item unumtags
-
-The number of unique tags found.
-
-=item utags
-
-An array reference containing zero or more
-L<RFID::Matrics::Tag|RFID::Matrics::Tag> objects representing the
-unique tags that were read.
-
-=back
-
-=cut
-
 sub readfullfield_unique
 {
     my $self = shift;
     my $pp = $self->readfullfield(@_);
 
-    @{$pp->{utags}} = uniq { tagcmp($a,$b) }
-                       sort { tagcmp($a,$b) } 
+    @{$pp->{utags}} = uniq { $a->tagcmp($b) }
+                       sort { $a->tagcmp($b) } 
                         @{$pp->{tags}};
     $pp->{unumtags}=scalar(@{$pp->{utags}});
     $pp;
 }
 
-
-=head3 start_constant_read
-
-Start a constant read until a I<stop_constant_read> command.  After
-calling this method, repeatedly call I<constant_read> to get tags that
-the reader sees.  The parameter to this method is a hash that can
-specify the I<node> and I<antenna> that should do the reading, and
-also:
-
-=over 4
-
-=item antenna1, antenna2, antenna3, antenna4
-
-Which antennas should be read in each time slot.  Leave a parameter
-undefined to skip that timeslot.  If I<antenna1> isn't given,
-I<antenna> is used.
-
-=item antenna1_power, antenna2_power, antenna3_power, antenna4_power
-
-The power level for the antenna in each of the time slots.
-
-=item dwell_time
-
-The number of milliseconds to scan for, between 6 and 150.
-
-=item channel
-
-The frequency channel to use, from 0 to 16.
-
-=back
-
-=cut
-
 sub start_constant_read
 {
     my $self = shift;
@@ -506,9 +693,9 @@ sub _cmd_start_constant_read
     my(%p)=@_;
 
     $self->_setdefaults(\%p);
-    $antflag{$_}=1
-	foreach grep { defined } ($p{antenna1}||$p{antenna}||MATRICS_ANT_1,
-				  @$p{qw(antenna2 antenna3 antenna4)});
+#    $antflag{$_}=1
+#	foreach grep { defined } ($p{antenna1}||$p{antenna}||MATRICS_ANT_1,
+#				  @$p{qw(antenna2 antenna3 antenna4)});
     $self->_makepacket(%p,
 		       cmd => 0x25,
 		       data => pack("CCCCCCCCCCCCa8",
@@ -531,47 +718,30 @@ sub _cmd_start_constant_read
 sub _epc_parsetags
 {
     my $self = shift;
-    my($count,$dc)=@_;
+    my($count,$dc,%tagprops)=@_;
     my @tags;
 
     foreach my $i (1..$count)
     {
-	my $type = unpack("C",substr($dc,0,1,""));
-	my $len = ($type == 0x0C) ? 8 : 12;
-	my $id_bits = unpack("a*",substr($dc,0,$len,""));
-	my $tag = RFID::Matrics::Tag->new(type => $type, id_bits => $id_bits)
-	    or die "Couldn't create new RFID::Matrics::Tag object: $!\n";
+	my $type = ord(substr($dc,0,1));
+	my $tag;
+	if ($type == 0x0C)
+	{
+	    # Proprietary Matrics Tag
+	    $tag = RFID::Matrics::Tag->new(id => bin2hex_little_endian(unpack("a8",substr($dc,1,8))),
+					   %tagprops);
+	}
+	else
+	{
+	    # EPC tag
+	    $tag = RFID::EPC::Tag->new(id => bin2hex_little_endian(unpack("a12",substr($dc,0,13))),
+				       %tagprops);
+	}
 	push(@tags,$tag);
     }
     @tags;
 }
 
-
-=head3 constant_read
-
-Returns tags from an ongoing constant read operation.  You call
-I<start_constant_read> to start reading, and I<stop_constant_read>
-when you're finished.
-
-Each time this method is called, it returns a hashref with the
-following attributes:
-
-=over 4
-
-=item numtags
-
-The number of tags found
-
-=item tags
-
-An array reference containing zero or more
-L<RFID::Matrics::Tag|RFID::Matrics::Tag> objects representing the tags
-that were read.
-
-=back
-
-=cut
-
 sub constant_read
 {
     my $self = shift;
@@ -618,29 +788,33 @@ sub _resp_constant_read
 sub _parsetags
 {
     my $self = shift;
-    my($count,$dc)=@_;
+    my($count,$dc,%tagprops)=@_;
     my @tags;
 
     foreach my $i (1..$count)
     {
 	my $type_len_bits = unpack("C",substr($dc,0,1,""));
-	my $len = ($type_len_bits & 0x04) ? 12 : 8;
-	my $type = ($type_len_bits & 0x03);
+	my $len = ($type_len_bits & 0x10) ? 12 : 8;
+	my $type = ($type_len_bits & 0x0f);
 	my $id_bits = unpack("a*",substr($dc,0,$len,""));
-	my $tag = RFID::Matrics::Tag->new(type => $type, id_bits => $id_bits)
-	    or die "Couldn't create new RFID::Matrics::Tag object: $!\n";
+	my $tag;
+	if ($type == 0)
+	{
+	    # EPC tag
+	    $tag = RFID::EPC::Tag->new(id => bin2hex_little_endian($id_bits),
+				       %tagprops);
+	}
+	else
+	{
+	    # Proprietary Matrics Tag
+	    $tag = RFID::Matrics::Tag->new(id => bin2hex_little_endian($id_bits),
+					   %tagprops);
+	}
 	push(@tags,$tag);
     }
     @tags;
 }
 
-
-=head3 stop_constant_read
-
-Take the reader out of constant read mode.
-
-=cut
-
 sub stop_constant_read
 {
     my $self = shift;
@@ -671,14 +845,6 @@ sub _resp_stop_constant_read
     return $pp;
 }
 
-
-=head3 stop_all_constant_read
-
-Stop all ongoing constant read operations, returning the reader to its
-default mode.
-
-=cut
-
 sub stop_all_constant_read
 {
     my $self = shift;
@@ -699,13 +865,16 @@ sub _make_confwhich_ant
     my($p)=@_;
     my %antflag;
 
-    foreach my $a (grep { defined } 
-		   ($p->{antenna1}||$p->{antenna}||MATRICS_ANT_1,
-		    @$p{qw(antenna2 antenna3 antenna4)}))
-    {
-	$antflag{$a}=1;
-    }
     
+    foreach my $a (grep { defined($p->{"antenna".$_}) }
+		   (1..4))
+    {
+	$antflag{$antid{$a}} = 1;
+    }
+    if (!(keys %antflag))
+    {
+	$antflag{$p->{antenna}||MATRICS_ANT_1} = 1;
+    }
     return ($antflag{MATRICS_ANT_1()}?1:0,
 	    $antflag{MATRICS_ANT_2()}?1:0,
 	    $antflag{MATRICS_ANT_3()}?1:0,
@@ -737,20 +906,15 @@ sub _make_combine_antenna_bits
     return ord($p->{combine_antenna_bits});
 }
 
-
-=head3 epc_readfullfield
-
-Reads all tags in the field of the specified antenna.  Parameters and
-return values are identical to I<readfullfield>.
-
-=cut
-
 sub epc_readfullfield
 {
     my $self = shift;
+    my(%p)=@_;
+    $self->_setdefaults(\%p);
+    
     my @resp = $self->_do_something(\&_cmd_epc_readfullfield,
 				    \&_resp_epc_readfullfield,
-				    @_)
+				    %p)
 	or return undef;
 
     my $ret = shift(@resp);
@@ -784,43 +948,23 @@ sub _resp_epc_readfullfield
     $pp;
 }
 
-
-=head3 epc_readfullfield_unique
-
-Performs an I<epc_readfullfield>, then adds attributes to the hash
-reference representing the I<unique> tags.  Pretty much the same as
-I<readfullfield_unique>.
-
-=cut
-
 sub epc_readfullfield_unique
 {
     my $self = shift;
     my $ret = $self->epc_readfullfield;
 
-    @{$ret->{utags}} = uniq { tagcmp($a,$b) }
-                       sort { tagcmp($a,$b) } 
+    @{$ret->{utags}} = uniq { $a->tagcmp($b) }
+                       sort { $a->tagcmp($b) } 
                         @{$ret->{tags}};
     $ret->{unumtags}=scalar(@{$ret->{utags}});
 
     $ret;
 }
 
-
-=head3 epc_getparamblock
-
-Get parameters for a particular reader and antenna using the EPC
-style.  The only difference seems to be the way filters are returned,
-and since filters are not yet implemented, this method is currently
-functionally identical to I<getparamblock>, although it sends a
-different command.  The parameters are identical.
-
-=cut
-
 sub epc_getparamblock
 {
     my $self = shift;
-    $self->_do_something(\&_cmd_getparamblock,\&_resp_getparamblock,@_);
+    $self->_do_something(\&_cmd_epc_getparamblock,\&_resp_epc_getparamblock,@_);
 }
 
 sub _cmd_epc_getparamblock
@@ -840,67 +984,48 @@ sub _resp_epc_getparamblock
     my $pp = $self->_parsepacket(@_)
 	or return undef;
     
-    @$pp{qw(power_level environment combine_antenna_bits protocol_speed filter_type reserved_bits filter_bits reserved_bits)} =
-	unpack("CCCCCa2a8a*",$pp->{data});
+    @$pp{qw(power_level environment combine_antenna_bits protocol_speed filter_type reserved1_bits filter_bits reserved2_bits)} =
+	unpack("CCCCCa3a8a16",$pp->{data});
     $pp;
 }
 
-
-=head3 epc_setparamblock
-
-Set parameters for a particular reader and antenna using the EPC
-style.  The only difference between this command and I<setparamblock>
-seems to be the way filters are defined, and since filters are not yet
-implemented, this method is currently functionally identical to
-I<setparamblock>, although it sends a different command.  The
-parameters are identical.
-
-=cut
-
 sub epc_setparamblock
 {
     my $self = shift;
-    $self->_do_something(\&_cmd_setparamblock,\&_resp_setparamblock,@_);
+    $self->_do_something(\&_cmd_epc_setparamblock,\&_resp_epc_setparamblock,@_);
 }
 
 sub _cmd_epc_setparamblock
 {
     my $self = shift;
     my(%p)=@_;
+    
 
     $self->_makepacket(%p,
 		       cmd => 0x15,
-		       data => pack("CCCCCCCCCx3a8x16",
-				    $self->_make_confwhich_ant(\%p),
-				    $p{power_level}||0xff,
+		       data => pack("CCCCCCCCCa3a8a16",
+				    $self->_make_confwhich_ant(\%p), # Returns 4 bytes
+				    defined($p{power_level})?$p{power_level}:0xff,
 				    $p{environment}||0x00,
 				    $self->_make_combine_antenna_bits(\%p),
+				    $p{protocol_speeed}||0x00,
 				    $p{filter_type}||0x00,
-				    $p{filter_bits}||"\0"x8,
+				    $p{reserved1_bits}||("\0"x3),
+				    $p{filter_bits}||("\0"x8),
+				    $p{reserved2_bits}||("\0"x16),
 				    )
 		       );
 }
 
 sub _resp_epc_setparamblock
 {
-    my $self;
+    my $self = shift;
+
     my $pp = $self->_parsepacket(@_)
 	or return undef;
 }
 
 
-=head3 epc_changeparamblock
-
-Modifies parameters for a particular reader and antenna using the EPC
-style.  This command is implemented in terms of I<epc_getparamblock>
-and I<epc_setparamblock>.  The only difference between this command
-and I<changeparamblock> seems to be the way filters are defined, and
-since filters are not yet implemented, this method is currently
-functionally identical to I<setparamblock>, although it sends
-different commands.  The parameters are identical.
-
-=cut
-
 sub epc_changeparamblock
 {
     my $self = shift;
@@ -913,33 +1038,6 @@ sub epc_changeparamblock
     return $self->epc_setparamblock(%$curparam, @_);
 }
 
-=head3 setnodeaddress
-
-Sets the node number for the reader with the given serial number.
-Takes a hash that can contain the following parameters:
-
-=over 4
-
-=item serialnum
-
-The serial number of the reader unit, as a hexadecimal string.  For
-example:
-
-  serialnum => "00000000000003AF"
-
-=item newnode
-
-The new node number.  If not given, I<node> is used.
-
-=item oldnode
-
-The old node number that the message should be addressed to.  If not
-given, the broadcast address C<0xFF> is used.
-
-=back
-
-=cut
-
 sub setnodeaddress
 {
     my $self = shift;
@@ -970,7 +1068,7 @@ sub _cmd_setnodeaddress
     if (!$p{serialnum_bits})
     {
 	defined($p{serialnum}) or return $self->error("Missing required parameter serialnum or serialnum_bits");
-	$p{serialnum_bits} = hex2bin($p{serialnum});
+	$p{serialnum_bits} = hex2bin_little_endian($p{serialnum});
     }
     $p{newnode} or $p{node} or return $self->error("Missing required parameter newnode or node");
 
@@ -991,29 +1089,6 @@ sub _resp_setnodeaddress
 	or return undef;
 }
 
-
-=head3 getreaderstatus
-
-Gets the status of a reader, including firmware version.  Takes as
-parameters a hash that can have the node and address.  Returns a hash
-reference with the following attributes:
-
-=over 4
-
-=item serialnum
-
-The serial number of the reader.
-
-=item version
-
-The firmware version number.
-
-=back
-
-Other attributes are not yet implemented.
-
-=cut
-
 sub getreaderstatus
 {
     my $self = shift;
@@ -1041,29 +1116,10 @@ sub _resp_getreaderstatus
 	    last_error reserved)} =
 		unpack("a8CCCCa4a4Ca11",$pp->{data});
     $pp->{version}=join(".",@$pp{qw(version_major version_minor version_eng)});
-    $pp->{serialnum} = bin2hex($pp->{serialnum_bits});
+    $pp->{serialnum} = bin2hex_little_endian($pp->{serialnum_bits});
     $pp;
 }
 
-
-=head3 getnodeaddress
-
-Gets the node address of the reader with the given serial number.
-Takes a hash containing the following parameters:
-
-=over 4
-
-=item serialnum
-
-A hex string of the serial number of the router.
-
-=back
-
-Returns a hash reference with a I<node> parameter giving the node
-number.
-
-=cut
-
 sub getnodeaddress
 {
     my $self = shift;
@@ -1079,7 +1135,7 @@ sub _cmd_getnodeaddress
     if (!$p{serialnum_bits})
     {
 	defined($p{serialnum}) or return $self->error("Missing required parameter serialnum or serialnum_bits");
-	$p{serialnum_bits} = hex2bin($p{serialnum});
+	$p{serialnum_bits} = hex2bin_little_endian($p{serialnum});
     }
     $self->_makepacket(%p,
 		      node => 0xff,
@@ -1155,36 +1211,6 @@ sub finish
 	or warn "Couldn't stop all constant readers: $!\n";
 }
 
-# Utility Functions
-
-sub sortcmp
-{
-    my $sub = shift;
-    local($a,$b)=@_;
-    $sub->();
-}
-
-sub hexdump
-{
-    my @a = split(//,$_[0]);
-    sprintf "%02x " x scalar(@a),map { ord } @a;
-}
-
-sub uniq(&@)
-{
-    my($cmpsub, @list)=@_;
-    my $last = shift @list
-	or return ();
-    my @ret =($last);
-    foreach (@list)
-    {
-	push(@ret,$_)
-	    unless sortcmp($cmpsub,$_,$last)==0;
-	$last = $_;
-    }
-    @ret;
-}
-
 sub error
 {
     my $self = shift;
@@ -1192,28 +1218,111 @@ sub error
 
     $self->{error}=$em;
     $self->{errcode}=defined($ec)?$ec:1;
-    warn "Error: $em\n"
-	if ($self->{debug});
+    $self->debug("Error: $em\n");
     return undef;
 }
 
 # Convert a hex string to binary, LSB first
-sub hex2bin
+sub hex2bin_little_endian
 {
     my $hex = $_[0];
     $hex =~ tr/0-9a-fA-F//cd;
     pack("C*",map { hex } reverse unpack("a2"x(length($hex)/2),$hex));
 }
 
-sub bin2hex
+sub hex2bin_big_endian
 {
-    my @a = split(//,$_[0]);
-    sprintf "%02x" x scalar(@a), reverse map {ord} @a;
+    my $hex = $_[0];
+    $hex =~ tr/0-9a-fA-F//cd;
+    pack("C*",map { hex } unpack("a2"x(length($hex)/2),$hex));
 }
 
+sub bin2hex_little_endian
+{
+    unpack("H*",pack("C*",reverse(unpack("C*",$_[0]))));
+}
+
+sub bin2hex_big_endian
+{
+    unpack("H*",$_[0]);
+#    my @a = split(//,$_[0]);
+#    sprintf "%02x" x scalar(@a), map {ord} @a;
+}
+
+=head2 Properties
+
+=head3 Antenna
+
+The default antenna for get operations; see also
+L<AntennaSequence|/AntennaSequence>.  This defaults to 1 if it is not
+set.
+
+=head3 AntennaSequence
+
+An arrayref of the antennas that should be queried, and in what order.
+The antenna names for a 4-port Matrics reader are simply 1, 2, 3, and
+4.  For example:
+
+    $reader->set(AntennaSequence => [1,2,3,4]);
+
+The default AntennaSequence is the L<default Antenna|/Antenna>.
+
+=head3 Debug
+
+Control the amount debugging information sent to C<STDERR>.  A higher
+value for this property will cause more information to be output.
+
+=head3 UniqueTags
+
+A boolean value controlling whether duplicate tags should be removed
+from the list returned by L<readtags|/readtags>.
+
+=head3 Environment
+
+How long an antenna should try to read tags during a
+L<readtags|readtags> command, between 0 and 4.  0 will read for a very
+short time, and is appropriate for environments where tags come and go
+very quickly, and it's OK if you miss a tag somtimes.  4 will read for
+longer, and is appropriate where tags stay relatively static and you
+want the reader to try its best to find all of them.
+
+When this item is retreived, you get the value for the default
+antenna.  When it's set, it's set for all of the antennas in the
+L<AntennaSequence|/AntennaSequence>.  To set the level for only one
+antenna, use C<Environment_AntennaI<n>>, where C<I<n>> is the number
+of the antenna you'd like to set.
+
+=head3 Node
+
+The Matrics node address associated with this object.  It defaults to
+4.
+
+=head3 PowerLevel
+
+The amount of power an antenna should use when doing a read, between 0
+and 255.  255 is full-power; the scale of this setting is logarithmic,
+so 208 is about 50% power, and 0x80 is about 25% power.
+
+When this item is retreived, you get the value for the default
+antenna.  When it's set, it's set for all of the antennas in the
+L<AntennaSequence|/AntennaSequence>.  To set the level for only one
+antenna, use C<PowerLevel_AntennaI<n>>, where C<I<n>> is the number of
+the antenna you'd like to set.
+
+=head3 ReaderVersion
+
+The software version running on this reader, as a string.  Cannot be
+L<set|/set>.
+
+=head3 ReaderSerialNum
+
+The serial number of this reader, as a string.  Cannot be L<set|/set>.
+
 =head1 SEE ALSO
 
-L<RFID::Matrics::Reader::Serial>, L<RFID::Matrics::Reader::TCP>, L<RFID::Matrics::Tag>.
+L<RFID::Matrics::Reader::Serial>, L<RFID::Matrics::Reader::TCP>,
+L<RFID::EPC::Tag>, L<RFID::Matrics::Tag>,
+L<http://www.eecs.umich.edu/~wherefid/code/rfid-perl/>.
 
 =head1 AUTHOR
 
